/*
 * @project        The CERN Tape Archive (CTA)
 * @copyright      Copyright(C) 2015-2021 CERN
 * @license        This program is free software: you can redistribute it and/or modify
 *                 it under the terms of the GNU General Public License as published by
 *                 the Free Software Foundation, either version 3 of the License, or
 *                 (at your option) any later version.
 *
 *                 This program is distributed in the hope that it will be useful,
 *                 but WITHOUT ANY WARRANTY; without even the implied warranty of
 *                 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *                 GNU General Public License for more details.
 *
 *                 You should have received a copy of the GNU General Public License
 *                 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <xroot_plugins/XrdCtaStream.hpp>
#include <xroot_plugins/XrdSsiCtaRequestMessage.hpp>
#include "version.h"

namespace cta { namespace xrd {

/*!
 * Stream object which implements "tapepool ls" command
 */
class VersionStream: public XrdCtaStream{
public:
  /*!
   * Constructor
   *
   * @param[in]    requestMsg    RequestMessage containing command-line arguments
   */
  VersionStream(const RequestMessage &requestMsg,cta::catalogue::Catalogue &catalogue, cta::Scheduler &scheduler, const std::string & catalogueConnString);
  
private:
    static constexpr const char* const LOG_SUFFIX  = "VersionStream";    //!< Identifier for log messages
    
    Versions m_client_versions;
    Versions m_server_versions;
    std::string m_catalogue_conn_string;
    std::string m_catalogue_version;
    
    bool m_is_done = false;
    
    /*!
    * Fill the buffer
    */
    virtual int fillBuffer(XrdSsiPb::OStreamBuffer<Data> *streambuf);
    
  /*!
  * Only one item is sent so isDone = true
  */
  virtual bool isDone() const {
    return m_is_done;
  }
  
};

VersionStream::VersionStream(const RequestMessage &requestMsg,cta::catalogue::Catalogue &catalogue, cta::Scheduler &scheduler,const std::string & catalogueConnString):
XrdCtaStream(catalogue,scheduler),
  m_client_versions(requestMsg.getClientVersions()),
  m_catalogue_conn_string(catalogueConnString),
  m_catalogue_version(m_catalogue.getSchemaVersion().getSchemaVersion<std::string>())
{
   XrdSsiPb::Log::Msg(XrdSsiPb::Log::DEBUG, LOG_SUFFIX, "VersionStream() constructor");
   m_server_versions.ctaVersion = CTA_VERSION;
   m_server_versions.xrootdSsiProtoIntVersion = XROOTD_SSI_PROTOBUF_INTERFACE_VERSION;
}

int VersionStream::fillBuffer(XrdSsiPb::OStreamBuffer<Data> *streambuf){
  m_is_done = true;
  Data record;
  auto version = record.mutable_version_item();
  auto client_version = version->mutable_client_version();
  client_version->set_cta_version(m_client_versions.ctaVersion);
  client_version->set_xrootd_ssi_protobuf_interface_version(m_client_versions.xrootdSsiProtoIntVersion);
  auto server_version = version->mutable_server_version();
  server_version->set_cta_version(m_server_versions.ctaVersion);
  server_version->set_xrootd_ssi_protobuf_interface_version(m_server_versions.xrootdSsiProtoIntVersion);
  version->set_catalogue_connection_string(m_catalogue_conn_string);
  version->set_catalogue_version(m_catalogue_version);
  streambuf->Push(record);
  
  return streambuf->Size();
}

}}